@@ -39,6 +39,7 @@ enum wpaevent {
   WPA_ACTIOND_EVENT_UNCHANGED,
   WPA_ACTIOND_EVENT_CONNECTED,
   WPA_ACTIOND_EVENT_DISCONNECTED,
+  WPA_ACTIOND_EVENT_FAILED,
   WPA_ACTIOND_EVENT_TERMINATED
 };
 
@@ -54,6 +55,7 @@ enum wpaaction {
   WPA_ACTIOND_ACTION_CONNECT,
   WPA_ACTIOND_ACTION_CONNECTION_LOST,
   WPA_ACTIOND_ACTION_CONNECTION_REESTABLISHED,
+  WPA_ACTIOND_ACTION_FAILED,
   WPA_ACTIOND_ACTION_DISCONNECT
 };
 
@@ -65,6 +67,7 @@ enum wpa_actiond_logevent {
   WPA_ACTIOND_LOG_CONNECTION_LOST,
   WPA_ACTIOND_LOG_CONNECTION_REESTABLISHED,
   WPA_ACTIOND_LOG_TERMINATE,
+  WPA_ACTIOND_LOG_FAILED,
   WPA_ACTIOND_LOG_ERROR,
   WPA_ACTIOND_LOG_CUSTOM_ERROR
 };
@@ -119,6 +122,12 @@ static void logevent(enum wpa_actiond_logevent l, const char *iface, const char
       else
         fprintf(stderr, "Terminating wpa_actiond session for interface '%s'\n", iface);
       break;
+    case WPA_ACTIOND_LOG_FAILED:
+      if(!tostderr)
+        syslog(LOG_NOTICE, "Interface '%s' connecting failed", iface);
+      else
+        fprintf(stderr, "Interface '%s' connecting failed\n", iface);
+      break;
     case WPA_ACTIOND_LOG_ERROR:
       if(!tostderr)
         syslog(LOG_ERR, "Error (%s): %s%m", iface, arg);
@@ -138,7 +147,7 @@ static int min(int i, int j) {
   return i < j ? i : j;
 }
 
-static void get_ssid(struct wpa_ctrl *ctrl, char *ssid) {
+static void get_status(struct wpa_ctrl *ctrl, char *ssid, char *idstr) {
   char reply[BUFLEN], *pos = &reply[0], *pos2;
   size_t reply_len = BUFLEN-1;
   *ssid = '\0';
@@ -160,6 +169,22 @@ static void get_ssid(struct wpa_ctrl *ctrl, char *ssid) {
     strncpy(ssid, pos, min(pos2-pos, SSIDLEN-1));
     ssid[min(pos2-pos, SSIDLEN-1)] = '\0';
   }
+
+  pos = &reply[0];
+
+  while(1) {
+    pos = strstr(pos, "id_str=");
+    if(pos == NULL || pos == reply || *(pos-1) == '\n')
+      break;
+    pos += strlen("id_str=");
+  }
+
+  if(pos != NULL) {
+    pos += strlen("id_str=");
+    pos2 = index(pos, '\n');
+    strncpy(idstr, pos, min(pos2-pos, IDSTRLEN-1));
+    idstr[min(pos2-pos, IDSTRLEN-1)] = '\0';
+  }
 }
 
 static int str_match(const char *s1, const char *s2) {
@@ -197,6 +222,8 @@ static void parse_wpa_event(const char *msg, enum wpaevent *ev, char *idstr)
     *ev = WPA_ACTIOND_EVENT_DISCONNECTED;
   else if (str_match(pos, WPA_EVENT_TERMINATING))
     *ev = WPA_ACTIOND_EVENT_TERMINATED;
+  else if (str_match(pos, WPA_EVENT_EAP_FAILURE))
+    *ev = WPA_ACTIOND_EVENT_FAILED;
 }
 
 static void action(enum wpaaction act, const char *iface, const char *ssid, const char *idstr, const int ctrlfd, const char *script) {
@@ -216,6 +243,9 @@ static void action(enum wpaaction act, const char *iface, const char *ssid, cons
     case WPA_ACTIOND_ACTION_DISCONNECT:
       actstr = "DISCONNECT";
       break;
+    case WPA_ACTIOND_ACTION_FAILED:
+      actstr = "FAILED";
+      break;
   }
 
   if(actstr == NULL)
@@ -255,6 +285,7 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
   int r;
   /* select timeout */
   struct timeval timeout;
+  struct timeval ping_timeout;
   /* save ssid */
   char ssid[SSIDLEN], old_ssid[SSIDLEN];
   char idstr[IDSTRLEN], old_idstr[IDSTRLEN];
@@ -305,7 +336,7 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
   }
 
   /* set initial ssid */
-  get_ssid(ctrl, ssid);
+  get_status(ctrl, ssid, idstr);
 
   logevent(WPA_ACTIOND_LOG_STARTED, iface, "");
 
@@ -314,8 +345,10 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
     FD_ZERO(&ctrl_fds);
     ctrl_fd = wpa_ctrl_get_fd(ctrl);
     FD_SET(ctrl_fd, &ctrl_fds);
+    ping_timeout.tv_sec = 5;
+    ping_timeout.tv_usec = 0;
     /* wait for an event */
-    r = select(ctrl_fd+1, &ctrl_fds, NULL, NULL, (state == WPA_ACTIOND_STATE_CONNECTION_LOST) ? &timeout : NULL);
+    r = select(ctrl_fd+1, &ctrl_fds, NULL, &ctrl_fds, (state == WPA_ACTIOND_STATE_CONNECTION_LOST) ? &timeout : &ping_timeout);
     switch(r) {
       case -1:
         /* select error, terminate wpa_actiond */
@@ -328,10 +361,26 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
         state = WPA_ACTIOND_STATE_TERMINATED;
         break;
       case 0:
-        /* timeout, we are in WPA_ACTIOND_STATE_CONNECTION_LOST, disconnect */
-        logevent(WPA_ACTIOND_LOG_DISCONNECTED, iface, ssid);
-        state = WPA_ACTIOND_STATE_DISCONNECTED;
-        action(WPA_ACTIOND_ACTION_DISCONNECT, iface, ssid, idstr, wpa_ctrl_get_fd(ctrl), script);
+        if (state == WPA_ACTIOND_STATE_CONNECTION_LOST) {
+          /* timeout, we are in WPA_ACTIOND_STATE_CONNECTION_LOST, disconnect */
+          logevent(WPA_ACTIOND_LOG_DISCONNECTED, iface, ssid);
+          state = WPA_ACTIOND_STATE_DISCONNECTED;
+          action(WPA_ACTIOND_ACTION_DISCONNECT, iface, ssid, idstr, wpa_ctrl_get_fd(ctrl), script);
+        } else {
+          /* ping wpa_supplicant if is still alive */
+          if(wpa_ctrl_request(ctrl, "PING", strlen("PING"), reply, &reply_len, NULL))
+            reply_len = 0;
+          reply[reply_len] = '\0';
+          if(!str_match(reply, "PONG")) {
+            /* supplicant has been terminated */
+            if(state == WPA_ACTIOND_STATE_CONNECTED || state == WPA_ACTIOND_STATE_CONNECTION_LOST) {
+              logevent(WPA_ACTIOND_LOG_DISCONNECTED, iface, ssid);
+              action(WPA_ACTIOND_ACTION_DISCONNECT, iface, ssid, idstr, wpa_ctrl_get_fd(ctrl), script);
+            }
+            logevent(WPA_ACTIOND_LOG_TERMINATE, iface, "");
+            state = WPA_ACTIOND_STATE_TERMINATED;
+          }
+        }
         break;
       default:
         /* event received */
@@ -341,8 +390,8 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
         parse_wpa_event(reply, &ev, idstr);
         switch(ev) {
           case WPA_ACTIOND_EVENT_CONNECTED:
-            /* connect to wireless */
-            get_ssid(ctrl, ssid);
+            /* connect to network */
+            get_status(ctrl, ssid, idstr);
             if(state == WPA_ACTIOND_STATE_CONNECTION_LOST) {
               if(strcmp(old_ssid, ssid)) {
                 logevent(WPA_ACTIOND_LOG_DISCONNECTED, iface, old_ssid);
@@ -378,6 +427,12 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
               timeout.tv_usec = 0;
             }
             break;
+          case WPA_ACTIOND_EVENT_FAILED:
+            /* connecting failed */
+            get_status(ctrl, ssid, idstr);
+            logevent(WPA_ACTIOND_LOG_FAILED, iface, ssid);
+            action(WPA_ACTIOND_ACTION_FAILED, iface, ssid, idstr, wpa_ctrl_get_fd(ctrl), script);
+            break;
           case WPA_ACTIOND_EVENT_TERMINATED:
             /* supplicant has been terminated */
             if(state == WPA_ACTIOND_STATE_CONNECTED || state == WPA_ACTIOND_STATE_CONNECTION_LOST) {
@@ -403,7 +458,7 @@ static void loop(const char *iface, const char *ctrlpath, const int disconnect_t
 static void usage(const char *progname) {
   fprintf(stderr, "\nUsage:\n"
                   "  %s -i interface -a script [-n] [-p control_socket_path] [-P pid_file] [-t timeout]\n\n"
-                  "  -i\tname of the wireless interface\n"
+                  "  -i\tname of the interface\n"
                   "  -a\taction script\n"
                   "  -n\tdo not daemonize, log to stderr (default: daemonize, log to syslog)\n"
                   "  -p\tpath to the wpa_supplicant control socket directory (default: %s)\n"
@@ -411,9 +466,9 @@ static void usage(const char *progname) {
                   "  -t\ttimeout in seconds until a lost connection is considered disconnected (default: %d)\n\n"
                   "  The action script takes four parameters:\n"
                   "    1) The interface name\n"
-                  "    2) The SSID of the network\n"
+                  "    2) The SSID of the wireless network or empty string if using wired driver\n"
                   "    3) The id_str parameter of the wpa_supplicant network section\n"
-                  "    4) One of the strings CONNECT, LOST, REESTABLISHED and DISCONNECT\n"
+                  "    4) One of the strings CONNECT, LOST, REESTABLISHED, FAILED and DISCONNECT\n"
                   "\n", progname, DEFAULT_CTRL_PATH, DEFAULT_PIDFILE, DEFAULT_DISCONNECT_TIMEOUT);
 }
 
